import {
	M3ReflectorName,
	m3Reflectors,
	M3RotorName,
	m3Rotots as m3Rotors,
	M4ReflectorName,
	m4Reflectors,
	M4RotorName,
	m4Rotors,
} from "./components";
import { Plugboard } from "./plugboard";
import { Rotor } from "./rotors";
import { checkCharacter } from "./validation";

export class Machine {
	plugboard: Plugboard;
	reflector: Map<string, string>;
	rotors: Rotor[];

	constructor(
		reflector: M3ReflectorName,
		first: M3RotorName,
		second: M3RotorName,
		third: M3RotorName,
	);
	constructor(
		reflector: M4ReflectorName,
		first: M3RotorName,
		second: M3RotorName,
		third: M3RotorName,
		fourth: M4RotorName,
	);
	constructor(
		reflector: M3ReflectorName | M4ReflectorName,
		first: M3RotorName,
		second: M3RotorName,
		third: M3RotorName,
		fourth?: M4RotorName,
	) {
		this.plugboard = new Plugboard();
		const reflectorComponent = getComponent(
			{ ...m3Reflectors, ...m4Reflectors },
			reflector,
		);

		this.reflector = new Map();
		for (let i = 0; i < 26; i++) {
			const aCharCode = "A".charCodeAt(0);
			this.reflector.set(
				String.fromCharCode(aCharCode + i),
				reflectorComponent.reflector[i],
			);
		}

		const rotorComponents = [
			getComponent(m3Rotors, first),
			getComponent(m3Rotors, second),
			getComponent(m3Rotors, third),
		];
		if (fourth) rotorComponents.push(getComponent(m4Rotors, fourth));

		this.rotors = rotorComponents.map(
			({ name, values, turnovers }) => new Rotor(name, values, turnovers),
		);
	}

	rings(): string[] {
		return this.rotors.map((r) => r.position());
	}

	stecker(one: string, two: string) {
		this.plugboard.stecker(one, two);
	}

	input(message: string): string {
		let output = "";
		for (let char of message) {
			checkCharacter(char, "character");
			let turnover = true;

			// There is a rare case when rotor three turns, and it does not cause rotor
			// two to turn (double-step): on the FIRST keypress when BOTH wheels are set
			// to turn. In that case, it's important not to set the second wheel as the
			// third one steps, since it will already have stepped and stepping twice in
			// one keypress is not possible mechanically.
			let rotorTwoStepped = false;

			// plugboard, first pass
			let scrambled = this.plugboard.in(char);

			// rotors, first pass (in)
			for (let i = 0; i < this.rotors.length; i++) {
				const rotor = this.rotors[i];
				if (turnover) {
					turnover = rotor.rotate();

					if (i === 1) {
						rotorTwoStepped = true;
					}

					// Because of a quirk in the machines design, if the third rotor
					// rotates, the second rotor also rotates. This funny action was
					// called "double-stepping". Although Naval Enigma added a fourth
					// rotor, it never causes the third rotor to double-step because
					// the fourth rotor never actually rotated. Rotor one is not
					// affected because it always rotates anyway.
					if (i === 2 && !rotorTwoStepped) {
						this.rotors[1].rotate();
					}
				}

				scrambled = rotor.in(scrambled);
			}

			// reflector
			scrambled = this.reflector.get(scrambled)!;

			// rotors, second pass (out)
			for (let i = this.rotors.length - 1; i >= 0; i--) {
				const rotor = this.rotors[i];
				scrambled = rotor.out(scrambled);
			}

			// plugboard, second pass
			output += this.plugboard.in(scrambled);

			let state = "";
			for (const r of this.rotors) {
				state += ` ${r.name}, ${r.offset}`;
			}
		}

		return output;
	}
}

function getComponent<T extends { name: string }>(
	components: Record<string, T>,
	componentName: string,
): T {
	const component = components[componentName];
	if (!component) throw new Error(`unknown component: ${component}`);

	return component;
}
